/*** SETTINGS ***/
//#define CONVERT_TO_16BPP
#define OUTPUT_FILE "/mnt/gbatmp/binkcap/protracker_output.mkv"
/*** END SETTINGS ***/

#include <dlfcn.h>
#include <fcntl.h>
#include <unistd.h>

#include <cstdio>
#include <vector>
#include <algorithm>

#include <sys/time.h>

#define CALL_OLD

#include "zmbv_dosbox.h"
#include "libmkv.h"
#include "rgbtorgb.hh"

#include <SDL/SDL_video.h>
#define SDL_UpperBlit_params \
             SDL_Surface *src, SDL_Rect *srcrect, \
             SDL_Surface *dst, SDL_Rect *dstrect

int SDLCALL SDL_UpperBlit( SDL_UpperBlit_params)
{
    /* This object will load the _real_ XPutImage
     * symbol and enable calling it
     */
    static struct caller
    {
        int (*orig) (SDL_UpperBlit_params);
    public:
        caller()
        {
            void *f = dlopen("libSDL-1.2.so.0", RTLD_NOW);
            orig = (int(*)(SDL_UpperBlit_params))dlsym(f, "SDL_UpperBlit");
            dlclose(f);
        }
    } caller;
    static struct sender
    {
        bool first;
        VideoCodec zmbv;
        size_t     bufsize;
        char*      buffer;
        unsigned   framecount;
        zmbv_format_t format;
        mk_Writer* mkw;
        mk_Track*  mkt;
        int64_t    first_time, last_time;
        unsigned   bytesperpixel;
        std::vector<unsigned char> FrameBuffer;
        std::vector<void*        > LinePointers;

        sender() : first(true), zmbv(), bufsize(0), buffer(0), framecount(0), format(ZMBV_FORMAT_NONE)
        {
        }
        ~sender()
        {
            Close();
        }
        void Close()
        {
            delete[] buffer;
            mk_close(mkw);
            fprintf(stderr, "Closing ZMBV & MKV\n");
        }
        void Blit(const SDL_Surface* src, const SDL_Rect* rect)
        {
            if(first)
            {
#ifdef CONVERT_TO_16BPP
                format        = ZMBV_FORMAT_16BPP;
                bytesperpixel = 2;
#else
                format        = zmbv.BPPFormat(src->format->BitsPerPixel);
                bytesperpixel = src->format->BytesPerPixel;
#endif

                fprintf(stderr, "SDL format: %ux%u, %u bpp,%u bytes per pixel,pitch=%u, R=%06X G=%06X B=%06X\n",
                    src->w,
                    src->h,
                    src->format->BitsPerPixel,
                    src->format->BytesPerPixel,
                    src->pitch,
                    src->format->Rmask,
                    src->format->Gmask,
                    src->format->Bmask);
                if(!zmbv.SetupCompress(src->w, src->h))
                {
                    fprintf(stderr, "setupcompress failed\n");
                }
                bufsize = zmbv.NeededSize(src->w, src->h, format);
                buffer  = new char[bufsize];
                fprintf(stderr, "zmbv buffer at %p, %zu bytes\n", buffer, bufsize);

                mk_TrackConfig mkc = { };
                mkc.trackType  = MK_TRACK_VIDEO;
                mkc.flagLacing = MK_LACING_NONE;
                mkc.flagEnabled = 1;
                mkc.flagDefault = 1;
                mkc.defaultDuration = 0;//(1/60.0) * 1e9;
                mkc.name = 0;
                mkc.codecID   = "V_MS/VFW/FOURCC";
                mkc.codecName = "ZMBV";
                char buf[40] = { 0 };
                memcpy(buf+16, CODEC_4CC, 4); // MPlayer looks for fourcc in this offset
                mkc.codecPrivate     = buf;
                mkc.codecPrivateSize = 40;    // MPlayer requires at least this much

                mkc.extra.video.pixelWidth = src->w;
                mkc.extra.video.pixelHeight = src->h;
                mkc.extra.video.displayWidth  = src->w;
                mkc.extra.video.displayHeight = src->h;
                mkc.extra.video.aspectRatioType = MK_ASPECTRATIO_FIXED;

                mkw = mk_createWriter(OUTPUT_FILE, 1000000, 1);
                mkt = mk_createTrack(mkw, &mkc);
                mk_writeHeader(mkw, "xpatch");

                FrameBuffer.resize( src->w * src->h * bytesperpixel );
                LinePointers.resize( src->h );
                for(unsigned y=0; y< src->h; ++y)
                    LinePointers[y] = (void*) &FrameBuffer[y * src->w * bytesperpixel];
            }
            int64_t cur_time;
            struct timeval tv;
            gettimeofday(&tv, NULL);
            cur_time = (tv.tv_usec + (int64_t)tv.tv_sec * 1000000) * 1000ll;
            if(first)
            {
                last_time = first_time = cur_time;
            }

            bool do_key = false;
            if(framecount%120 == 0 || (cur_time - last_time) > 1000000000ll)
            {
                do_key=true;
            }

            if(!zmbv.PrepareCompressFrame( do_key, format, 0, buffer, bufsize))
            {
                fprintf(stderr, "preparecompressframe failed\n");
            }

            unsigned sbpp = src->format->BytesPerPixel;
            char* p = (char*) src->pixels;
            for(unsigned yo=0; yo< rect->h; ++yo)
            {
                unsigned y = rect->y + yo;
                unsigned x = rect->x;
                unsigned nx = rect->w;
                char *q = (char*) &p[ y * src->pitch + x * sbpp ];
                char* t = (char*) &FrameBuffer[ (y * src->w + x) * bytesperpixel ];
#ifdef CONVERT_TO_16BPP
                if(sbpp == 2)
                    memcpy(t, q, nx*2);
                else if(sbpp == 4)
                    Convert32ToR16Frame(q, (unsigned char*) t, nx, nx, y);
                else
                    Convert24ToR16Frame(q, (unsigned char*) t, nx, nx, y);
#else
                memcpy(t, q, nx * bytesperpixel);
#endif
            }

            zmbv.CompressLines(src->h, &LinePointers[0]);

            size_t gotsize = zmbv.FinishCompressFrame();

            fprintf(stderr, "Frame %u at time=%llu, %u byte pixels, zmbv format %u, key=%s; size = %zu\n",
                framecount, (unsigned long long) (cur_time - first_time),
                bytesperpixel, format,
                do_key?"true":"false",
                gotsize);

            mk_startFrame(mkw, mkt);
            mk_addFrameData(mkw, mkt, buffer, gotsize);
            mk_setFrameFlags(mkw, mkt, cur_time - first_time, do_key, 0); //(1/60.0) * 1e9);
            mk_flushFrame(mkw, mkt);


            ++framecount;
            first    = false;
            last_time = cur_time;
        }
    } sender;
    if(!dst) { sender.Close(); return 0; }
    int res = 0;
#ifdef CALL_OLD
    res = caller.orig(src,srcrect,dst,dstrect);
#endif
    sender.Blit(src, srcrect);
    return res;
}

void SDL_Quit()
{
    static struct caller
    {
        void (*orig) (void);
    public:
        caller()
        {
            void *f = dlopen("libSDL-1.2.so.0", RTLD_NOW);
            orig = (void(*)(void))dlsym(f, "SDL_Quit");
            dlclose(f);
        }
    } caller;
    SDL_UpperBlit(0,0,0,0);
    caller.orig();
}
